///
/// Copyright 2006 Attrice Corporation, http://www.attrice.info 
/// This software is a freeware and may be used under Attrice Sidekicks License
/// (download updated license from http://www.attrice.info/downloads/sidekick_license.txt)
/// Feedback or questions? Contact us at info@attrice.info
///
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Diagnostics;

using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.VersionControl.Client;

namespace Attrice.VisualStudioIntegration.Tfs
{
    /// <summary>
    /// Class encapsulating all TFS functionality to checkout/checkin build type files
    /// </summary>
    class BuildTypeEditor
    {
        #region Member variables
        private TeamFoundationServer _serverTfs;
        private VersionControlServer _serverVc;
        private string _projectName;
        private string _buildTypeName;
        #endregion

        #region Member properties
        string UserName
        {
            get
            {
                return _serverTfs.AuthenticatedUserName;
            }
        }

        string ProjectName
        {
            get
            {
                return _projectName;
            }
        }

        string ProjectServerPath
        {
            get
            {
                return String.Format("$/{0}", ProjectName);
            }
        }

        string BuildTypeName
        {
            get
            {
                return _buildTypeName;
            }
        }

        string BuildTypeServerPath
        {
            get
            {
                return String.Format("$/{0}/TeamBuildTypes/{1}", ProjectName, BuildTypeName);
            }
        }

        string TemporaryPath
        {
            get
            {
                return Path.GetTempPath();
            }
        }

        string ProjectLocalPath
        {
            get
            {
                return Path.Combine(TemporaryPath, ProjectName);
            }
        }

        string BuildTypeLocalPath
        {
            get
            {
                return Path.Combine(Path.Combine(ProjectLocalPath, "TeamBuildTypes"),BuildTypeName);
            }
        }

        string ComputerName
        {
            get
            {
                return Workstation.Current.Name;
            }
        }

        string WorkspaceName
        {
            get
            {
                return "BuildTypesSidekick";
            }
        }

        string WorkspaceComment
        {
            get
            {
                return "Build Types Sidekick workspace. Please do not modify manually.";
            }
        }

        string CheckinComment
        {
            get
            {
                return "Build Types Sidekick edit";
            }
        }
        #endregion

        public string BuildTypeFileName
        {
            get
            {
                return Path.Combine(BuildTypeLocalPath, "TFSBuild.proj");
            }
        }

        public string TfsServerName
        {
            get
            {
                return _serverTfs.Name;
            }
        }

        /// <summary>
        /// Creates build type editor and initializes TFS services
        /// </summary>
        /// <param name="tfsServerName">TFS server name</param>
        public BuildTypeEditor(string tfsServerName)
        {
            try
            {
                _serverTfs = TeamFoundationServerFactory.GetServer(tfsServerName);
                _serverVc = _serverTfs.GetService(typeof(VersionControlServer)) as VersionControlServer;
            }
            catch (Exception e)
            {
                throw new ApplicationException(String.Format("Failed to initialize (server {0}). {1}", tfsServerName, e.Message), e);
            }
        }

        /// <summary>
        /// Selects project and build type for editing
        /// </summary>
        /// <param name="projectName">Selected project name</param>
        /// <param name="buildTypeName">Selected build type name</param>
        public void Select(string projectName, string buildTypeName)
        {
            _projectName = projectName;
            _buildTypeName = buildTypeName;
        }

        [Flags]
        public enum BuildTypeSccStatus
        {
            Undefined = 0,            
            CheckedOut = 1,
            CheckedOutLocally = 2,
            CheckedIn = 4
        }

        /// <summary>
        /// Determines the build types status 
        /// </summary>
        /// <returns>Build types SCC status</returns>
        public BuildTypeSccStatus GetStatus()
        {
            Debug.Assert(_projectName != null && _buildTypeName != null);
            try
            {
                // get workspace for local build files editing
                Workspace[] workspaces = _serverVc.QueryWorkspaces(WorkspaceName, UserName, ComputerName);
                PendingSet[] sets;
                if (workspaces.Length != 1)
                {
                    // check global files status
                    sets = _serverVc.QueryPendingSets(new string[] { BuildTypeServerPath }, RecursionType.OneLevel, null, null);
                }
                else
                {
                    // check files status in workspace
                    sets = workspaces[0].QueryPendingSets(new string[] { BuildTypeServerPath }, RecursionType.OneLevel, WorkspaceName, UserName, false);
                }
                // if there is at least one pending set, files are checked out either locally or somewhere else
                if (sets.Length > 0)
                {
                    // checked out remotely
                    if (workspaces.Length != 1)
                        return BuildTypeSccStatus.CheckedOut;
                    // checked out locally
                    return BuildTypeSccStatus.CheckedOutLocally;
                }
                // not checked out
                return BuildTypeSccStatus.CheckedIn;
            }
            catch (Exception e)
            {
                throw new ApplicationException(String.Format("Failed version control operation ({0})", e.Message), e);
            }
        }

        /// <summary>
        /// Perform get and check out for build type files; creates workspaces and mappings
        /// as needed
        /// </summary>
        /// <returns>True if operation suceeds; false if there was some problem with source control 
        /// operations</returns>
        /// <exception cref="ApplicationException">When TFS throws exception</exception>
        public bool Edit()
        {
            Debug.Assert(_projectName != null && _buildTypeName != null);
            try
            {
                // get workspace for local build files editing
                Workspace[] workspaces = _serverVc.QueryWorkspaces(WorkspaceName, UserName, ComputerName);                
                Workspace workspace = null;
                Debug.Assert(workspaces.Length < 2);
                // either create or take found workspace
                if (workspaces.Length == 0)
                    workspace = _serverVc.CreateWorkspace(WorkspaceName, UserName, WorkspaceComment);
                else
                    workspace = workspaces[0];
                // check - if project is not mapped, then map it
                if (!workspace.IsServerPathMapped(ProjectServerPath))
                    workspace.Map(ProjectServerPath, ProjectLocalPath);
                // Check if the files are not already checked in in workspace
                PendingSet[] sets = workspace.QueryPendingSets(new string[] { BuildTypeServerPath}, RecursionType.OneLevel, WorkspaceName, UserName, false);
                if (sets.Length == 1 && sets[0].PendingChanges.Length > 0)
                    return true;
                // Check if get may be performed without problems
                GetRequest request = new GetRequest(BuildTypeServerPath, RecursionType.OneLevel, VersionSpec.Latest);
                GetStatus status = workspace.Get(request, GetOptions.Preview | GetOptions.GetAll);
                // if status is not empty - will not proceed and let the user fix it
                if (!status.NoActionNeeded && (status.NumFailures > 0 || status.NumConflicts > 0 || status.NumWarnings > 0))
                    return false;
                // get the files latest version locally
                status = workspace.Get(request, GetOptions.GetAll);
                // if failed - let the user fix it
                if (status.NumFailures != 0)
                    return false;
                // pend edit with exclusive lock
                int result = workspace.PendEdit(new string[] { BuildTypeServerPath },
                    RecursionType.OneLevel, null, LockLevel.CheckOut);
                return (result >= 3);
            }
            catch (Exception e)
            {
                throw new ApplicationException(String.Format("Failed version control operation ({0})", e.Message), e);
            }
        }

        /// <summary>
        /// Performs check in or undo of build types edit
        /// </summary>
        /// <param name="undo">Specifies whether to perform check in or undo</param>
        public void CommitEdit(bool undo)
        {
            // find build type editing workspace
            Workspace[] workspaces =_serverVc.QueryWorkspaces(WorkspaceName, UserName, ComputerName);
            if (workspaces.Length != 1)
                throw new ApplicationException(String.Format("Workspace {0} not found", WorkspaceName));
            // check that server path mapping exists
            Workspace workspace = workspaces[0];
            if (!workspace.IsServerPathMapped(BuildTypeServerPath))
                throw new ApplicationException(String.Format("Path {0} is not mapped in {1}", BuildTypeServerPath, WorkspaceName));
            // get pending changes
            PendingSet[] sets = workspace.QueryPendingSets(new string[] { BuildTypeServerPath }, RecursionType.OneLevel, WorkspaceName, UserName, false);
            if (sets.Length != 1)
                throw new ApplicationException(String.Format("No pending edit changes in path {0}, mapped in {1}", BuildTypeServerPath, WorkspaceName));
            // undo or check in changes
            int result;
            if (undo)
                result = workspace.Undo(sets[0].PendingChanges); 
            else
                result = workspace.CheckIn(sets[0].PendingChanges, CheckinComment);
            // If some (or all) build type files are not changed, check in will succeed 
            // but return value will not be equal to three. So commenting out it for now
            // as do not know easy way to validate result
            /*if (result < 3)
                throw new ApplicationException("Failed to check in/undo pending changes");
             */ 
        }
    }
}
