using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Globalization;
using Microsoft.TeamFoundation.VersionControl.Client;

namespace Attrice.VisualStudioIntegration.Tfs
{
    public partial class FormSearchItem : Form
    {

        private VersionControlServer _versionControl;
        private DataTable _dtProjects;
        private string _serverItem;
        private bool _showDeleted;

        public string SelectedServerItem
        {
            get { return _serverItem; }
        }


        public FormSearchItem()
        {
            InitializeComponent();
            listViewItems.ListViewItemSorter = new CustomListSorter();
            buttonJump.Enabled = false;
        }

        private DataTable GetProjects()
        {
            DataTable table = new DataTable("projects");
            table.Columns.Add(ListTable.ColumnID);
            table.Columns.Add(ListTable.ColumnValue);
            TeamProject[] projects = _versionControl.GetAllTeamProjects(false);
            foreach (TeamProject project in projects)
            {
                DataRow dr = table.NewRow();
                dr[ListTable.ColumnID] = project.ServerItem;
                dr[ListTable.ColumnValue] = project.Name;
                table.Rows.Add(dr);
            }
            return table;
        }

        public void Initialize(VersionControlServer versionControl, 
            string selectedServerPath,
            bool showDeleted)
        {
            _versionControl = versionControl;
            _dtProjects = GetProjects();

            comboBoxSearchType.Items.Add("Files and folders");
            comboBoxSearchType.Items.Add("Files only");
            comboBoxSearchType.Items.Add("Folders only");
            comboBoxSearchType.SelectedIndex = 0;
            
            DataRow dr = _dtProjects.NewRow();
            dr[ListTable.ColumnValue] = "All projects";
            dr[ListTable.ColumnID] = "$/";
            _dtProjects.Rows.Add(dr);
                       
            DataView dv = new DataView(_dtProjects);
            dv.Sort = ListTable.ColumnID;
            comboBoxPath.DataSource = dv;
            comboBoxPath.DisplayMember = ListTable.ColumnValue;
            comboBoxPath.ValueMember = ListTable.ColumnID;

            // set deleted status
            _showDeleted = showDeleted;
            if (_showDeleted)
                listViewItems.Columns.Add("Deleted status", 60);

            // select default project
            if (!String.IsNullOrEmpty(selectedServerPath))
            {
                int index = selectedServerPath.IndexOf('/', 2);
                if (index != -1)
                {
                    comboBoxPath.SelectedValue = selectedServerPath.Substring(0, index);
                    textBoxSearchString.Text = selectedServerPath.Substring(index);
                }
                else
                    comboBoxPath.SelectedValue = selectedServerPath;
            }
            else
                comboBoxPath.SelectedValue = "$/";
        }

        private void buttonSearch_Click(object sender, EventArgs e)
        {
            ItemType itemType = ItemType.Any;
            if (comboBoxSearchType.SelectedIndex == 1)
                itemType = ItemType.File;
            if (comboBoxSearchType.SelectedIndex == 2)
                itemType = ItemType.Folder;

            string searchPath = comboBoxPath.SelectedValue.ToString();
            if (searchPath.IndexOf("$") == -1)
                searchPath = "$/" + searchPath;
            string searchString = "*";

            if (textBoxSearchString.TextLength > 0)
                searchString = textBoxSearchString.Text;

            if (searchPath[searchPath.Length - 1] != '/')
                searchPath += "/";
            searchPath += searchString;

            listViewItems.Items.Clear();
            ItemSet items = _versionControl.GetItems(searchPath, 
                LatestVersionSpec.Latest, 
                RecursionType.Full,
                _showDeleted?DeletedState.Any:DeletedState.NonDeleted, 
                itemType, false);
            foreach (Item item in items.Items)
            {
                ListViewItem lvi = new ListViewItem(Utilities.ParseServerItem(item.ServerItem));
                lvi.SubItems.Add(item.ServerItem.Substring(0, item.ServerItem.Length - lvi.Text.Length));
                lvi.SubItems.Add(item.ItemType.ToString());
                if (_showDeleted)
                    lvi.SubItems.Add((item.DeletionId > 0) ? "Deleted" : "Not deleted");
                lvi.Tag = item;
                listViewItems.Items.Add(lvi);
            }
        }

        private void buttonCancel_Click(object sender, EventArgs e)
        {
            DialogResult = DialogResult.Cancel;
            Close();
        }

        private void buttonJump_Click(object sender, EventArgs e)
        {
            if (listViewItems.SelectedIndices.Count != 1)
                return;

            DialogResult = DialogResult.OK;
            _serverItem = (listViewItems.SelectedItems[0].Tag as Item).ServerItem;
            Close();            
        }

        private void listViewItems_ColumnClick(object sender, ColumnClickEventArgs e)
        {
            ListView listView = (sender as ListView);
            CustomListSorter sorter = listView.ListViewItemSorter as CustomListSorter;
            sorter.SetColumn(listView.Columns[e.Column]);
            listView.Sort();
        }

        private void FormSearchItem_Activated(object sender, EventArgs e)
        {
            textBoxSearchString.Focus();
        }

        private void listViewItems_SelectedIndexChanged(object sender, EventArgs e)
        {
            buttonJump.Enabled = listViewItems.SelectedIndices.Count == 1;
        }       
    }

    public sealed class ListTable
    {
        public static readonly string ColumnID = "ID";
        public static readonly string ColumnValue = "Value";
    }

    public class Utilities
    {
        private static readonly string cDateTimeFormat = "dd-MMM-yyyy HH:mm:ss";

        public static bool ParseDateTimeInvariant(string value, out DateTime date)
        {
            return DateTime.TryParseExact(value, cDateTimeFormat, DateTimeFormatInfo.InvariantInfo, DateTimeStyles.AssumeLocal, out date);
        }

        public static string FormatDateTimeInvariant(DateTime date)
        {
            return date.ToString(cDateTimeFormat, DateTimeFormatInfo.InvariantInfo);
        }
                
        public static String ParseServerItem(String serverItem)
        {
            if (serverItem == "$/")
                return serverItem;
            int index = serverItem.LastIndexOf('/');
            if (index == -1)
                return String.Empty;
            return serverItem.Substring(index + 1);
        }
    }

    [Flags]
    public enum ListSorterType
    {
        Undefined = 0,
        Date = 1,
        Integer = 2
    }

    public class CustomListSorter : System.Collections.IComparer
    {
        private ColumnHeader _columnHeader;
        private ListSorterType _type;
        private bool _ascending;

        public bool Ascending
        {
            get { return _ascending; }
            set { _ascending = value; }
        }

        public CustomListSorter()
        {
            _ascending = false;
            _columnHeader = null;
            _type = ListSorterType.Undefined;
        }

        public CustomListSorter(ListSorterType type)
        {
            _ascending = false;
            _columnHeader = null;
            _type = type;
        }


        public void SetColumn(ColumnHeader columnHeader)
        {
            _columnHeader = columnHeader;
            if (_columnHeader.Tag == null)
                _columnHeader.Tag = _ascending;
            else
                _columnHeader.Tag = !((bool)(_columnHeader.Tag));
        }

        public int Compare(object x, object y)
        {
            bool bDescending = !_ascending;
            if (_columnHeader != null && _columnHeader.Tag != null)
                bDescending = (bool)(_columnHeader.Tag);
            ListViewItem lv1 = x as ListViewItem;
            ListViewItem lv2 = y as ListViewItem;
            int index = _columnHeader == null ? 0 : _columnHeader.Index;
            String string1 = lv1.SubItems[index].Text;
            String string2 = lv2.SubItems[index].Text;

            if ((int)(_type & ListSorterType.Date) > 0)
            {
                DateTime date1, date2;
                if (Utilities.ParseDateTimeInvariant(string1, out date1)
                    && Utilities.ParseDateTimeInvariant(string2, out date2))
                {
                    if (!bDescending)
                        return DateTime.Compare(date1, date2);
                    return DateTime.Compare(date2, date1);
                }
            }

            if ((int)(_type & ListSorterType.Integer) > 0)
            {
                Int32 value1, value2;
                if (Int32.TryParse(string1, out value1) && Int32.TryParse(string2, out value2))
                {
                    if (!bDescending)
                        return value1 - value2;
                    return value2 - value1;
                }
            }

            if (!bDescending)
                return String.Compare(string1, string2);
            return String.Compare(string2, string1);

        }
    }
}