///
/// Copyright 2006 Attrice Corporation, http://www.attrice.info 
/// This software is a freeware and may be used under Attrice Sidekicks License
/// (download updated license from http://www.attrice.info/downloads/sidekick_license.txt)
/// Feedback or questions? Contact us at info@attrice.info
///
using System;
using Extensibility;
using EnvDTE;
using EnvDTE80;
using Microsoft.VisualStudio.CommandBars;
using System.Resources;
using System.Reflection;
using System.Globalization;
using System.Diagnostics;
using System.Runtime.InteropServices;

using Microsoft.TeamFoundation.Common;
using Microsoft.VisualStudio.TeamFoundation;
using Microsoft.VisualStudio.Shell.Interop;
using OleInterop = Microsoft.VisualStudio.OLE.Interop;

namespace Attrice.VisualStudioIntegration.Tfs
{
    /// <summary>
    /// Main class advanced functionality (mainly related to team explorer selection)
    /// </summary>
	public partial class AddinMain : IDTExtensibility2, IDTCommandTarget
	{
        private static bool CompareCommand(string commandName, string shortCommandName)
        {
            return (commandName == "Attrice.VisualStudioIntegration.Tfs.AddinMain." + shortCommandName);
        }

        /// <summary>
        /// Adds named command to VS and to specified bar (performs various checks to avoid
        /// adding it several times)
        /// </summary>
        /// <param name="barName"></param>
        /// <param name="commandName"></param>
        /// <param name="commandTitle"></param>
        /// <param name="commandTooltip"></param>
        /// <param name="imageIndex"></param>
        /// <returns></returns>
        private CommandBarControl AddCommand(string barName, string commandName, 
            string commandTitle, string commandTooltip,int imageIndex)
        {
            object[] contextGUIDS = new object[] { };
            string fullCommandName = String.Format("{0}.{1}", "Attrice.VisualStudioIntegration.Tfs.AddinMain", commandName);
            Commands2 commands = (Commands2)_applicationObject.Commands;
            CommandBars bars = (CommandBars)_applicationObject.CommandBars;
            bool bCommandFound = false;
            // check that command does not exists
            for (int i = 1; i <= commands.Count; i++)
            {
                if (fullCommandName.CompareTo(commands.Item(i,0).Name) == 0)
                {
                    bCommandFound = true;
                    break;
                }                
            }
            Command command = null;
            // if command not found create it
            if (!bCommandFound)
            {
                command = commands.AddNamedCommand2(_addInInstance,
                    commandName, commandTitle,
                    commandTooltip, true, imageIndex, ref contextGUIDS,
                    (int)vsCommandStatus.vsCommandStatusInvisible,
                    (int)vsCommandStyle.vsCommandStylePictAndText,
                    vsCommandControlType.vsCommandControlTypeButton);
            }
            else
            {
                command = commands.Item(fullCommandName, 0);
            }
            if (command == null)
                throw new ApplicationException(String.Format("Failed to create command {0}", commandName)); ;
            // add command to bar
            CommandBar bar = bars[barName];
            if (bar == null)
                throw new ApplicationException(String.Format("Failed to find bar {0}", barName)); ;
            bool bControlFound = false;
            for (int i = 1; i <= bar.Controls.Count; i++)
                if (bar.Controls[i].Caption == commandTitle)
                    return bar.Controls[i];
            return command.AddControl(bar, 1) as CommandBarControl;
        }

        /// <summary>
        /// Adds separator before the specified command on the bar
        /// </summary>
        /// <param name="barName"></param>
        /// <param name="commandTitle"></param>
        private void AddSeparator(string barName, string commandTitle)
        {
            CommandBars bars = (CommandBars)_applicationObject.CommandBars;
            CommandBar bar = bars[barName];
            if (bar == null)
                throw new ApplicationException(String.Format("Failed to find bar {0}", barName)); ;
            for (int i = 1; i <= bar.Controls.Count; i++)
                if (bar.Controls[i].Caption == commandTitle)
                {
                    bar.Controls[i].BeginGroup = true;
                    return;
                }
            throw new ApplicationException(String.Format("Failed to find command {0}", commandTitle)); ;
        }

        /// <summary>
        /// Get selected TFS server, project and build type
        /// </summary>
        /// <param name="serverName"></param>
        /// <param name="projectName"></param>
        /// <param name="buildTypeName"></param>
        /// <returns></returns>
        private bool GetTfsSelected(out string serverName, out string projectName, out string buildTypeName)
        {
            serverName = null;
            projectName = null;
            buildTypeName = null;

            // get Team Explorer service
            OleInterop.IServiceProvider serviceProvider = (OleInterop.IServiceProvider)_applicationObject;
            Guid guid = new Guid(typeof(IVsTeamExplorer).GUID.ToString());
            IntPtr ptr;
            int result = serviceProvider.QueryService(ref guid, ref guid, out ptr);
            if (result != 0)
                throw new ApplicationException("Failed to get Team Explorer service");
            IVsTeamExplorer explorer = (IVsTeamExplorer)Marshal.GetObjectForIUnknown(ptr);            
            // explorer is not open yet
            if (explorer.TeamExplorerUIHierarchy == null)
                return false;
            
            // get TFS server node
            object value;            
            result = explorer.TeamExplorerUIHierarchy.GetProperty((uint)Microsoft.VisualStudio.VSConstants.VSITEMID_ROOT,
                (int)__VSHPROPID.VSHPROPID_Name, out value);
            // if failed to get hierarchy root            
            if (result != 0)
                throw new ApplicationException("Failed to get TFS server node name");
            serverName = value as string;
            
            // get project name
            ProjectContext projectContext = explorer.GetProjectContext();
            // if no selected project
            if (projectContext.ProjectName == null)
                return false;
            projectName = projectContext.ProjectName;
            
            // get selected item id
            uint itemID;
            IVsMultiItemSelect multiSelect;
            result = explorer.TeamExplorerWindow.GetCurrentSelection(out ptr, out itemID, out multiSelect);
            // if no selection
            if (result != 0)
                return false;
            
            // get selected node name (we suppose that menu is visible only on builds)
            IVsHierarchy hierarchy = (IVsHierarchy)Marshal.GetObjectForIUnknown(ptr);
            result = hierarchy.GetProperty(itemID, (int)__VSHPROPID.VSHPROPID_Name, out value);
            if (result != 0)
                return false;
            buildTypeName = value as string;            
            return true;
        }
        
        /// <summary>
        /// Get build type editor for operation on build types
        /// </summary>
        /// <returns>Valid initialized editor object or null</returns>
        private BuildTypeEditor GetBuildTypeEditor()
        {
            string serverName, projectName, buildTypeName;
            // check the values selected in team explorer
            if (!GetTfsSelected(out serverName, out projectName, out buildTypeName))
                return null;
            // make sure that we are working with the same TFS server (in improbable case that user changes server on the run)
            if (_editor != null && _editor.TfsServerName != serverName)
                _editor = null;
            if (_editor == null)
                _editor = new BuildTypeEditor(serverName);
            _editor.Select(projectName, buildTypeName);
            return _editor;
        }
    }
}
